import pytest
from celery.result import AsyncResult
from celery.states import ALL_STATES

from kedro_graphql.events import PipelineEventMonitor


@pytest.mark.usefixtures('mock_celery_session_app')
@pytest.mark.usefixtures('celery_session_worker')
@pytest.mark.usefixtures('depends_on_current_app')
class TestPipelineEventMonitor:

    @pytest.mark.asyncio
    async def test_consume_default(self, mocker, mock_celery_session_app, mock_app, mock_pipeline):
        """
        Requires Redis to run.
        """
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.before_start")
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.on_success")
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.on_retry")
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.on_failure")
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.after_return")

        async for e in PipelineEventMonitor(app=mock_celery_session_app, task_id=mock_pipeline.status[-1].task_id).start():
            print(e)
            assert e["status"] in ALL_STATES

    @pytest.mark.asyncio
    async def test_consume_short_timeout(self, mocker, mock_celery_session_app, mock_app, mock_pipeline):
        """
        Requires Redis to run.

        Test with shorter timeout to test Exception handling
        """
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.before_start")
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.on_success")
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.on_retry")
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.on_failure")
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.after_return")

        async for e in PipelineEventMonitor(app=mock_celery_session_app, task_id=mock_pipeline.status[-1].task_id, timeout=0.01).start():
            print(e)
            assert e["status"] in ALL_STATES

    @pytest.mark.asyncio
    async def test_consume_exception(self, mocker, mock_celery_session_app, mock_app, mock_pipeline):
        """
        Requires Redis to run.

        Let task finish before starting monitor test Exception handling
        """
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.before_start")
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.on_success")
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.on_retry")
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.on_failure")
        mocker.patch("kedro_graphql.tasks.KedroGraphqlTask.after_return")

        result = AsyncResult(mock_pipeline.status[-1].task_id).wait()

        async for e in PipelineEventMonitor(app=mock_celery_session_app, task_id=mock_pipeline.status[-1].task_id).start():
            print(e)
            assert e["status"] in ALL_STATES



"""
This is a boilerplate pipeline 'example01'
generated using Kedro 0.19.11
"""
import time


def uppercase(text: str) -> str:
    """Converts text to uppercase."""
    return text.upper()


def reverse(text: str) -> str:
    """Reverses the given text."""
    return text[::-1]


def append_timestamp(text: str) -> str:
    """Appends a timestamp to the text."""
    return f"{text} - {time.time()}"



import pytest
from celery.result import AsyncResult

from .conftest import IN_DEV, REASON


@pytest.mark.skipif(IN_DEV, reason=REASON)
@pytest.mark.usefixtures('celery_session_app')
@pytest.mark.usefixtures('celery_session_worker')
class TestArgoWorkflowRunner:
    @pytest.mark.asyncio
    async def test_runner(self, mock_pipeline_argo):
        """
        """
        # wait for task to finish
        r = AsyncResult(mock_pipeline_argo.task_id).get()

        # fetch result
        r = AsyncResult(mock_pipeline_argo.task_id)
        assert r.status == "SUCCESS"



import json

import pytest

from kedro_graphql.models import (
    DataSet,
    Parameter,
    Pipeline,
    PipelineStatus,
    State,
    Tag,
)
from kedro_graphql.tasks import run_pipeline


@pytest.mark.usefixtures('mock_celery_session_app')
@pytest.mark.usefixtures('celery_session_worker')
@pytest.mark.usefixtures('depends_on_current_app')
@pytest.mark.asyncio
async def test_run_pipeline(mock_app, mock_text_in, mock_text_out):
    """
    This test will fail because the pipeline is missing in the backend
    """
    inputs = [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})}]
    outputs = [{"name": "text_out", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_out)})}]
    parameters = [{"name": "example", "value": "hello"}]
    tags = [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]

    p = Pipeline(
        name="example00",
        data_catalog=[DataSet(**i) for i in inputs] + [DataSet(**o) for o in outputs],
        parameters=[Parameter(**p) for p in parameters],
        tags=[Tag(**p) for p in tags],
        status=[PipelineStatus(state=State.STAGED,
                               runner="kedro.runner.SequentialRunner",
                               session=None,
                               started_at=None,
                               finished_at=None,
                               task_id=None,
                               task_name=None)]
    )

    p = mock_app.backend.create(p)
    serial = p.serialize()

    result = run_pipeline.delay(
        id=str(p.id),
        name=serial["name"],
        data_catalog=serial["data_catalog"],
        parameters=serial["parameters"],
        runner="kedro.runner.SequentialRunner"
    )
    result = result.wait(timeout=None, interval=0.5)



import asyncio
from base64 import b64decode, b64encode
from datetime import datetime
from importlib import import_module
from typing import AsyncGenerator, Optional

import strawberry
from bson.objectid import ObjectId
from celery.states import UNREADY_STATES
from fastapi.encoders import jsonable_encoder
from kedro.framework.project import pipelines
from strawberry.tools import merge_types
from strawberry.types import Info

from . import __version__ as kedro_graphql_version
from .config import config
from .events import PipelineEventMonitor
from .hooks import InvalidPipeline
from .logs.logger import PipelineLogStream, logger
from .models import (
    PageMeta,
    Pipeline,
    PipelineEvent,
    PipelineInput,
    PipelineLogMessage,
    Pipelines,
    PipelineStatus,
    PipelineTemplate,
    PipelineTemplates,
    State,
)
from .tasks import run_pipeline


def encode_cursor(id: int) -> str:
    """
    Encodes the given id into a cursor.

    :param id: The ID to encode.

    :return: The encoded cursor.
    """
    return b64encode(f"cursor:{id}".encode("ascii")).decode("ascii")


def decode_cursor(cursor: str) -> int:
    """
    Decodes the ID from the given cursor.

    :param cursor: The cursor to decode.

    :return: The decoded user ID.
    """
    cursor_data = b64decode(cursor.encode("ascii")).decode("ascii")
    return cursor_data.split(":")[1]


@strawberry.type
class Query:

    @strawberry.field(description="Get a pipeline template.")
    def pipeline_template(self, info: Info, id: str) -> PipelineTemplate:
        for p in info.context["request"].app.kedro_pipelines_index:
            print(p.id, type(p.id))
            if p.id == id:
                return p

    @strawberry.field(description="Get a list of pipeline templates.")
    def pipeline_templates(self, info: Info, limit: int, cursor: Optional[str] = None) -> PipelineTemplates:
        if cursor is not None:
            # decode the user ID from the given cursor.
            pipe_id = ObjectId(decode_cursor(cursor=cursor))
        else:
            pipe_id = ObjectId("100000000000000000000000")  # unix epoch Jan 1, 1970 as objectId

        # filter the pipeline template data, going through the next set of results.
        filtered_data = [pipe for pipe in info.context["request"].app.kedro_pipelines_index
                         if pipe.id.generation_time >= pipe_id.generation_time]

        # slice the relevant pipeline template data (Here, we also slice an
        # additional pipe instance, to prepare the next cursor).
        sliced_pipes = filtered_data[: limit + 1]

        if len(sliced_pipes) > limit:
            # calculate the client's next cursor.
            last_pipe = sliced_pipes.pop(-1)
            next_cursor = encode_cursor(id=last_pipe.id)
        else:
            # We have reached the last page, and
            # don't have the next cursor.
            next_cursor = None

        return PipelineTemplates(
            pipeline_templates=sliced_pipes, page_meta=PageMeta(next_cursor=next_cursor)
        )

    @strawberry.field(description="Get a pipeline instance.")
    def read_pipeline(self, id: str, info: Info) -> Pipeline:
        try:
            p = info.context["request"].app.backend.read(id=id)
            if p is None:
                raise InvalidPipeline(f"Pipeline {id} does not exist in the project.")
        except Exception as e:
            raise InvalidPipeline(f"Error retrieving pipeline {id}: {e}")

        p.kedro_pipelines_index = info.context["request"].app.kedro_pipelines_index
        return p

    @strawberry.field(description="Get a list of pipeline instances.")
    def read_pipelines(self, info: Info, limit: int, cursor: Optional[str] = None, filter: Optional[str] = "",
                       sort: Optional[str] = "") -> Pipelines:
        if cursor is not None:
            # decode the user ID from the given cursor.
            pipe_id = decode_cursor(cursor=cursor)
        else:
            pipe_id = "000000000000000000000000"  # unix epoch Jan 1, 1970 as objectId

        results = info.context["request"].app.backend.list(cursor=pipe_id, limit=limit + 1, filter=filter, sort=sort)

        if len(results) > limit:
            # calculate the client's next cursor.
            last_pipe = results.pop(-1)
            next_cursor = encode_cursor(id=last_pipe.id)
        else:
            # We have reached the last page, and
            # don't have the next cursor.
            next_cursor = None

        return Pipelines(
            pipelines=results, page_meta=PageMeta(next_cursor=next_cursor)
        )


@strawberry.type
class Mutation:
    @strawberry.mutation(description="Execute a pipeline.")
    def create_pipeline(self, pipeline: PipelineInput, info: Info) -> Pipeline:
        """
        - is validation against template needed, e.g. check DataSet type or at least check dataset names
        """

        if pipeline.name not in pipelines.keys():
            raise InvalidPipeline(f"Pipeline {pipeline.name} does not exist in the project.")

        d = jsonable_encoder(pipeline)
        p = Pipeline.decode(d)

        serial = p.encode(encoder="kedro")

        # credentials not supported yet
        # merge any credentials with inputs and outputs
        # credentials are intentionally not persisted
        # NOTE celery result may persist creds in task result?

        started_at = datetime.now()
        p.created_at = started_at

        # Get kedro project, kedro-graphql, and pipeline versions
        p.project_version = config.get("KEDRO_PROJECT_VERSION", None)
        p.kedro_graphql_version = kedro_graphql_version
        p.pipeline_version = None
        package_name = config.get("KEDRO_PROJECT_NAME", None)
        if package_name:
            try:
                module = import_module(f".pipelines.{pipeline.name}", package=package_name)
                p.pipeline_version = getattr(module, "__version__", None)
            except Exception as e:
                logger.info(f"Could not find pipeline version: {e}")

        if d["state"] == "STAGED":
            p.status.append(PipelineStatus(state=State.STAGED,
                                           runner=None,
                                           session=None,
                                           started_at=None,
                                           finished_at=None,
                                           task_id=None,
                                           task_name=None))
            logger.info(f'Staging pipeline {p.name}')
            p = info.context["request"].app.backend.create(p)
            p.kedro_pipelines_index = info.context["request"].app.kedro_pipelines_index
            return p
        else:
            p.status.append(PipelineStatus(state=State.READY,
                                           runner=info.context["request"].app.config["KEDRO_GRAPHQL_RUNNER"],
                                           session=None,
                                           started_at=started_at,
                                           finished_at=None,
                                           task_id=None,
                                           task_name=str(run_pipeline)))

            p = info.context["request"].app.backend.create(p)

            result = run_pipeline.delay(
                id=str(p.id),
                name=serial["name"],
                parameters=serial["parameters"],
                data_catalog=serial["data_catalog"],
                runner=info.context["request"].app.config["KEDRO_GRAPHQL_RUNNER"],
                slices=d.get("slices", None),
                only_missing=d.get("only_missing", False)
            )

            logger.info(f'Running {p.name} pipeline with task_id: ' + str(result.task_id))
            p.kedro_pipelines_index = info.context["request"].app.kedro_pipelines_index
            return p

    @strawberry.mutation(description="Update a pipeline.")
    def update_pipeline(self, id: str, pipeline: PipelineInput, info: Info) -> Pipeline:

        try:
            p = info.context["request"].app.backend.read(id=id)
            if p is None:
                raise InvalidPipeline(f"Pipeline {id} does not exist in the project.")
        except Exception as e:
            raise InvalidPipeline(f"Error retrieving pipeline {id}: {e}")

        pipeline_input_dict = jsonable_encoder(pipeline)

        # Update pipeline with new pipeline input
        p.parameters = pipeline_input_dict.get("parameters")
        p.data_catalog = pipeline_input_dict.get("data_catalog")
        p.tags = pipeline_input_dict.get("tags")
        p.parent = pipeline_input_dict.get("parent")
        runner = pipeline_input_dict.get("runner") if pipeline_input_dict.get("runner") else config["KEDRO_GRAPHQL_RUNNER"]

        # If PipelineInput is READY and pipeline is not already running
        if pipeline_input_dict.get("state", None) == "READY" and p.status[-1].state.value not in UNREADY_STATES.union(["READY"]):

            if (p.status[-1].state.value != "STAGED"):
                # Add new status object to pipeline because this is another run attempt
                p.status.append(PipelineStatus(state=State.READY,
                                               runner=runner,
                                               session=None,
                                               started_at=datetime.now(),
                                               finished_at=None,
                                               task_id=None,
                                               task_name=str(run_pipeline)))
            else:
                # Replace staged status with running status
                p.status[-1] = PipelineStatus(state=State.READY,
                                              runner=runner,
                                              session=None,
                                              started_at=datetime.now(),
                                              finished_at=None,
                                              task_id=None,
                                              task_name=str(run_pipeline))

            # Update pipeline in backend before running task
            p = info.context["request"].app.backend.update(p)

            serial = p.encode(encoder="kedro")

            result = run_pipeline.delay(
                id=str(p.id),
                name=serial["name"],
                parameters=serial["parameters"],
                data_catalog=serial["data_catalog"],
                runner=runner,
                slices=pipeline_input_dict.get("slices", None),
                only_missing=pipeline_input_dict.get("only_missing", False)
            )

            logger.info(f'Running {p.name} pipeline with task_id: ' + str(result.task_id))

        # If PipelineInput is STAGED and pipeline is not already running or staged
        if pipeline_input_dict.get("state", None) == "STAGED" and p.status[-1].state.value not in UNREADY_STATES.union(["READY"]) and p.status[-1].state.value != "STAGED":
            p.status.append(PipelineStatus(state=State.STAGED,
                                           runner=runner,
                                           session=None,
                                           started_at=None,
                                           finished_at=None,
                                           task_id=None,
                                           task_name=None))
            logger.info(f'Staging pipeline {p.name}')
        p = info.context["request"].app.backend.update(p)

        p.kedro_pipelines_index = info.context["request"].app.kedro_pipelines_index
        return p

    @strawberry.mutation(description="Delete a pipeline.")
    def delete_pipeline(self, id: str, info: Info) -> Optional[Pipeline]:
        try:
            p = info.context["request"].app.backend.read(id=id)
            if p is None:
                raise InvalidPipeline(f"Pipeline {id} does not exist in the project.")
        except Exception as e:
            raise InvalidPipeline(f"Error retrieving pipeline {id}: {e}")

        info.context["request"].app.backend.delete(id=id)
        logger.info(f'Deleted {p.name} pipeline with id: ' + str(id))
        p.kedro_pipelines_index = info.context["request"].app.kedro_pipelines_index
        return p


@strawberry.type
class Subscription:
    @strawberry.subscription(description="Subscribe to pipeline events.")
    async def pipeline(self, id: str, info: Info, interval: float = 0.5) -> AsyncGenerator[PipelineEvent, None]:
        """Subscribe to pipeline events.
        """
        try:
            p = info.context["request"].app.backend.read(id=id)
            if p is None:
                raise InvalidPipeline(f"Pipeline {id} does not exist in the project.")
        except Exception as e:
            raise InvalidPipeline(f"Error retrieving pipeline {id}: {e}")

        while (not p.status[-1].task_id):
            # Wait for the task to be assigned a task_id
            await asyncio.sleep(0.1)
            p = info.context["request"].app.backend.read(id=id)

        if p:
            async for e in PipelineEventMonitor(app=info.context["request"].app.celery_app, task_id=p.status[-1].task_id).start(interval=interval):
                e["id"] = id
                yield PipelineEvent(**e)

    @strawberry.subscription(description="Subscribe to pipeline logs.")
    async def pipeline_logs(self, id: str, info: Info) -> AsyncGenerator[PipelineLogMessage, None]:
        """Subscribe to pipeline logs."""
        try:
            p = info.context["request"].app.backend.read(id=id)
            if p is None:
                raise InvalidPipeline(f"Pipeline {id} does not exist in the project.")
        except Exception as e:
            raise InvalidPipeline(f"Error retrieving pipeline {id}: {e}")

        while (not p.status[-1].task_id):
            # Wait for the task to be assigned a task_id
            await asyncio.sleep(0.1)
            p = info.context["request"].app.backend.read(id=id)

        if p:
            stream = await PipelineLogStream().create(task_id=p.status[-1].task_id, broker_url=info.context["request"].app.config["KEDRO_GRAPHQL_BROKER"])
            async for e in stream.consume():
                e["id"] = id
                yield PipelineLogMessage(**e)


def build_schema(type_plugins):
    ComboQuery = merge_types("Query", tuple([Query] + type_plugins["query"]))
    ComboMutation = merge_types("Mutation", tuple([Mutation] + type_plugins["mutation"]))
    ComboSubscription = merge_types("Subscription", tuple([Subscription] + type_plugins["subscription"]))

    return strawberry.Schema(query=ComboQuery, mutation=ComboMutation, subscription=ComboSubscription)



import pathlib
from importlib import import_module

import click
import uvicorn
from kedro.framework.session import KedroSession
from kedro.framework.startup import bootstrap_project
from watchfiles import run_process

from .config import config
from .logs.logger import logger


def init_app(app, config, session):
    # create app instance
    module, class_name = app.rsplit(".", 1)
    module = import_module(module)
    class_inst = getattr(module, class_name)
    return class_inst(kedro_session=session, config=config)


def start_app(app, config, conf_source, env, package_name, project_path):
    bootstrap_project(project_path)
    with KedroSession.create(project_path=project_path, env=env, conf_source=conf_source) as session:
        a = init_app(app, config, session)
        uvicorn.run(a, host="0.0.0.0", port=5000, log_level="info")


def start_worker(app, config, conf_source, env, package_name, project_path):

    bootstrap_project(project_path)
    with KedroSession.create(project_path=project_path, env=env, conf_source=conf_source) as session:
        a = init_app(app, config, session)
        from .celeryapp import celery_app
        capp = celery_app(a.config, a.backend)
        worker = capp.Worker()
        worker.start()


@click.group(name="kedro-graphql")
def commands():
    pass


@commands.command()
@click.pass_obj
@click.option(
    "--app",
    "-a",
    default=config["KEDRO_GRAPHQL_APP"],
    help="Application import path"
)
@click.option(
    "--backend",
    default=config["KEDRO_GRAPHQL_BACKEND"],
    help="The only supported value for this option is 'kedro_graphql.backends.mongodb.MongoBackend'"
)
@click.option(
    "--broker",
    default=config["KEDRO_GRAPHQL_BROKER"],
    help="URI to broker e.g. 'redis://localhost'"
)
@click.option(
    "--celery-result-backend",
    default=config["KEDRO_GRAPHQL_CELERY_RESULT_BACKEND"],
    help="URI to backend for celery results e.g. 'redis://localhost'"
)
@click.option(
    "--conf-source",
    default=config["KEDRO_GRAPHQL_CONF_SOURCE"],
    help="Path of a directory where project configuration is stored."
)
@click.option(
    "--env",
    "-e",
    default=config["KEDRO_GRAPHQL_ENV"],
    help="Kedro configuration environment name. Defaults to `local`."
)
@click.option(
    "--imports",
    "-i",
    default=config["KEDRO_GRAPHQL_IMPORTS"],
    help="Additional import paths"
)
@click.option(
    "--mongo-uri",
    default=config["MONGO_URI"],
    help="URI to mongodb e.g. 'mongodb://root:example@localhost:27017/'"
)
@click.option(
    "--mongo-db-name",
    default=config["MONGO_DB_NAME"],
    help="Name to use for collection in mongo e.g. 'pipelines'"
)
@click.option(
    "--runner",
    default=config["KEDRO_GRAPHQL_RUNNER"],
    help="Execution mechanism to run pipelines e.g. 'kedro.runner.SequentialRunner'"
)
@click.option(
    "--log-tmp-dir",
    default=config["KEDRO_GRAPHQL_LOG_TMP_DIR"],
    help="Temporary directory for logs"
)
@click.option(
    "--log-path-prefix",
    default=config["KEDRO_GRAPHQL_LOG_PATH_PREFIX"],
    help="Prefix of path to save logs"
)
@click.option(
    "--reload",
    "-r",
    is_flag=True,
    default=False,
    help="Enable auto-reload."
)
@click.option(
    "--reload-path",
    default=None,
    type=click.Path(exists=True, resolve_path=True, path_type=pathlib.Path),
    help="Path to watch for file changes, defaults to <project path>/src"
)
@click.option(
    "--worker",
    "-w",
    is_flag=True,
    default=False,
    help="Start a celery worker."
)
def gql(metadata, app, backend, broker, celery_result_backend, conf_source,
        env, imports, mongo_uri, mongo_db_name, runner, log_tmp_dir, log_path_prefix, reload, reload_path, worker):
    """Commands for working with kedro-graphql."""

    config.update({
        "MONGO_URI": mongo_uri,
        "MONGO_DB_NAME": mongo_db_name,
        "KEDRO_GRAPHQL_IMPORTS": imports,
        "KEDRO_GRAPHQL_APP": app,
        "KEDRO_GRAPHQL_BACKEND": backend,
        "KEDRO_GRAPHQL_BROKER": broker,
        "KEDRO_GRAPHQL_CELERY_RESULT_BACKEND": celery_result_backend,
        "KEDRO_GRAPHQL_RUNNER": runner,
        "KEDRO_GRAPHQL_ENV": env,
        "KEDRO_GRAPHQL_CONF_SOURCE": conf_source,
        "KEDRO_GRAPHQL_LOG_TMP_DIR": log_tmp_dir,
        "KEDRO_GRAPHQL_LOG_PATH_PREFIX": log_path_prefix,
        "KEDRO_PROJECT_VERSION": getattr(import_module(metadata.package_name), "__version__", None),
        "KEDRO_PROJECT_NAME": metadata.package_name
    })

    if not reload_path:
        reload_path = metadata.project_path.joinpath("src")

    if reload:
        logger.info("AUTO-RELOAD ACTIVATED, watching '" + str(reload_path) + "' for changes")

    if worker:
        if reload:
            run_process(str(reload_path), target=start_worker, args=(
                app, config, conf_source, env, metadata.package_name, metadata.project_path))
        else:
            start_worker(app, config, conf_source, env, metadata.package_name, metadata.project_path)

    else:
        if reload:
            run_process(reload_path, target=start_app, args=(app, config, conf_source,
                        env, metadata.package_name, metadata.project_path))
        else:
            start_app(app, config, conf_source, env, metadata.package_name, metadata.project_path)



import pytest

from kedro_graphql.logs.logger import PipelineLogStream


@pytest.mark.usefixtures('mock_celery_session_app')
@pytest.mark.usefixtures('celery_session_worker')
@pytest.mark.usefixtures('depends_on_current_app')
class TestPipelineLogStream:
    @pytest.mark.asyncio
    async def test_consume(self, mock_app, mock_pipeline):
        """Requires Redis to run.
        """
        task_id = mock_pipeline.status[-1].task_id
        subscriber = await PipelineLogStream().create(task_id=task_id, broker_url=mock_app.config["KEDRO_GRAPHQL_BROKER"])
        async for e in subscriber.consume():
            assert set(e.keys()) == set(["task_id", "message_id", "message", "time"])



import json
import logging
import os
import shutil
from datetime import date, datetime
from pathlib import Path
from typing import Dict, List

from celery import Task, shared_task
from kedro import __version__ as kedro_version
from kedro.framework.project import pipelines
from kedro.framework.session import KedroSession
from kedro.io import AbstractDataset, DataCatalog
from omegaconf import OmegaConf

from kedro_graphql.logs.logger import KedroGraphQLLogHandler
from kedro_graphql.runners import init_runner

from .config import config as CONFIG
from .models import DataSet, State

logger = logging.getLogger("kedro")

class KedroGraphqlTask(Task):

    _db = None

    @property
    def db(self):
        if self._db is None:
            self._db = self.app.kedro_graphql_backend
        return self._db


    def before_start(self, task_id, args, kwargs):
        """Handler called before the task starts.

        .. versionadded:: 5.2

        Arguments:
            task_id (str): Unique id of the task to execute.
            args (Tuple): Original arguments for the task to execute.
            kwargs (Dict): Original keyword arguments for the task to execute.

        Returns:
            None: The return value of this handler is ignored.
        """
        handler = KedroGraphQLLogHandler(task_id, broker_url = self._app.conf["broker_url"])
        logger.addHandler(handler)

        p = self.db.read(id=kwargs["id"])
        p.status[-1].state = State.STARTED
        p.status[-1].task_id = task_id
        p.status[-1].task_args = json.dumps(args)
        p.status[-1].task_kwargs = json.dumps(kwargs)
        self.db.update(p)

        try:
            # Create info and error handlers for the run
            os.makedirs(os.path.join(CONFIG["KEDRO_GRAPHQL_LOG_TMP_DIR"], task_id), exist_ok=True)
            formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
            info_handler = logging.FileHandler(os.path.join(CONFIG["KEDRO_GRAPHQL_LOG_TMP_DIR"], task_id + '/info.log'), 'a')
            info_handler.setLevel(logging.INFO)
            info_handler.setFormatter(formatter)
            error_handler = logging.FileHandler(os.path.join(CONFIG["KEDRO_GRAPHQL_LOG_TMP_DIR"], task_id + '/errors.log'), 'a') 
            error_handler.setLevel(logging.ERROR)
            error_handler.setFormatter(formatter)
            logger.addHandler(info_handler)
            logger.addHandler(error_handler)
            logger.info(f"Storing tmp logs in {os.path.join(CONFIG['KEDRO_GRAPHQL_LOG_TMP_DIR'], task_id)}")
            
            # Ensure KEDRO_GRAPHQL_LOG_PATH_PREFIX is provided
            log_path_prefix = CONFIG.get('KEDRO_GRAPHQL_LOG_PATH_PREFIX')
            if log_path_prefix:

                today = date.today()

                # Add metadata and log datasets to data catalog
                gql_meta = DataSet(name="gql_meta", config=json.dumps({"type": "json.JSONDataset",
                                                            "filepath": os.path.join(log_path_prefix,f"year={today.year}",f"month={today.month}",f"day={today.day}",str(p.id),"meta.json")}))
                gql_logs = DataSet(name="gql_logs",config=json.dumps({"type": "partitions.PartitionedDataset",
                                                        "dataset": "text.TextDataset",
                                                        "path": os.path.join(log_path_prefix,f"year={today.year}",f"month={today.month}",f"day={today.day}",str(p.id))}))
                p.data_catalog.append(gql_meta)
                p.data_catalog.append(gql_logs)

                # Save metadata to S3
                AbstractDataset.from_config(gql_meta.name, json.loads(gql_meta.config)).save(p.serialize())
                p = self.db.update(p)

                logger.info(f"Capturing pipeline metadata in {os.path.join(log_path_prefix,f'year={today.year}',f'month={today.month}',f'day={today.day}',str(p.id),'meta.json')}")
                logger.info(f"Capturing pipeline logs in {os.path.join(log_path_prefix,f'year={today.year}',f'month={today.month}',f'day={today.day}',str(p.id))}")

                # Capture pipeline object returned as an attribute of the task object
                setattr(self, "kedro_graphql_pipeline", p)
            else:
                logger.info(f"Missing KEDRO_GRAPHQL_LOG_PATH_PREFIX in config. Not capturing session logs.")
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")


    def on_success(self, retval, task_id, args, kwargs):
        """Success handler.

        Run by the worker if the task executes successfully.

        Arguments:
            retval (Any): The return value of the task.
            task_id (str): Unique id of the executed task.
            args (Tuple): Original arguments for the executed task.
            kwargs (Dict): Original keyword arguments for the executed task.

        Returns:
            None: The return value of this handler is ignored.
        """
        
        p = self.db.read(id=kwargs["id"])
        p.status[-1].state = State.SUCCESS
        self.db.update(p)


    def on_retry(self, exc, task_id, args, kwargs, einfo):
        """Retry handler.

        This is run by the worker when the task is to be retried.

        Arguments:
            exc (Exception): The exception sent to :meth:`retry`.
            task_id (str): Unique id of the retried task.
            args (Tuple): Original arguments for the retried task.
            kwargs (Dict): Original keyword arguments for the retried task.
            einfo (~billiard.einfo.ExceptionInfo): Exception information.

        Returns:
            None: The return value of this handler is ignored.
        """
        
        p = self.db.read(id=kwargs["id"])
        p.status[-1].state = State.RETRY
        p.status[-1].task_exception = str(exc)
        p.status[-1].task_einfo = str(einfo)
        self.db.update(p)        
   

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        """Error handler.

        This is run by the worker when the task fails.

        Arguments:
            exc (Exception): The exception raised by the task.
            task_id (str): Unique id of the failed task.
            args (Tuple): Original arguments for the task that failed.
            kwargs (Dict): Original keyword arguments for the task that failed.
            einfo (~billiard.einfo.ExceptionInfo): Exception information.

        Returns:
            None: The return value of this handler is ignored.
        """

        p = self.db.read(id=kwargs["id"])
        p.status[-1].state = State.FAILURE
        p.status[-1].task_exception = str(exc)
        p.status[-1].task_einfo = str(einfo)
        self.db.update(p)


    def after_return(self, status, retval, task_id, args, kwargs, einfo):
        """Handler called after the task returns.

        Arguments:
            status (str): Current task state.
            retval (Any): Task return value/exception.
            task_id (str): Unique id of the task.
            args (Tuple): Original arguments for the task.
            kwargs (Dict): Original keyword arguments for the task.
            einfo (~billiard.einfo.ExceptionInfo): Exception information.

        Returns:
            None: The return value of this handler is ignored.
        """

        finished_at = datetime.now()

        p = self.db.read(id=kwargs["id"])
        p.status[-1].finished_at = finished_at
        p.status[-1].task_result = str(retval)
        self.db.update(p)

        logger.info("Closing log stream")

        # Clean up log handlers
        for handler in logger.handlers:
            if isinstance(handler, KedroGraphQLLogHandler) and handler.topic == task_id:
                handler.flush()
                handler.close()
                handler.broker.connection.delete(task_id) ## delete stream
                handler.broker.connection.close()
            if isinstance(handler, logging.FileHandler):
                handler.close()
        logger.handlers = []

        # Clear logs from temp_logs
        try:
            shutil.rmtree(os.path.join(CONFIG["KEDRO_GRAPHQL_LOG_TMP_DIR"], task_id))  # Removes the directory and its contents
        except Exception as e:
            logger.info(f"Failed to clear logs in {os.path.join(CONFIG['KEDRO_GRAPHQL_LOG_TMP_DIR'].name, task_id)}: {e}")


@shared_task(bind = True, base = KedroGraphqlTask)
def run_pipeline(self, 
                 id: str = None,
                 name: str = None, 
                 parameters: dict = None, 
                 data_catalog: dict = None,
                 runner: str = None,
                 slices: List[Dict[str, List[str]]] = None,
                 only_missing: bool = False):
    
    with KedroSession.create(project_path = Path(__file__).resolve().parent.parent.parent,
                            env = CONFIG["KEDRO_GRAPHQL_ENV"],
                            conf_source = CONFIG["KEDRO_GRAPHQL_CONF_SOURCE"]) as session:
        
        hook_manager = session._hook_manager
        
        p = self.db.read(id=id)
        p.status[-1].session = session.session_id
        self.db.update(p)

        # If modified data catalog object with gql_meta and gql_logs datasets exists, use it
        if getattr(self, "kedro_graphql_pipeline", None):
            logger.info("using data_catalog with gql_meta and gql_logs")
            serial = self.kedro_graphql_pipeline.serialize()
            catalog = {**serial["data_catalog"], **data_catalog}
        else:
            logger.info("using data_catalog parameter to build data catalog")
            catalog = data_catalog

        io = DataCatalog().from_config(catalog=catalog)

        ## add parameters to DataCatalog using OmegaConf and dotlist notation
        parameters_dotlist = [f"{key}={value}" for key, value in parameters.items()]
        conf_parameters = OmegaConf.from_dotlist(parameters_dotlist)
        io.add_feed_dict({"parameters": conf_parameters})

        params_dotlist = [f"params:{key}={value}" for key, value in parameters.items()]
        conf_params = OmegaConf.from_dotlist(params_dotlist)
        io.add_feed_dict(conf_params)

        try:
            hook_manager.hook.after_catalog_created(
                catalog=io,
                conf_catalog=None,
                conf_creds=None,
                feed_dict=None,
                save_version=None,
                load_versions=None
            )
            # Populate the filtering parameters based on the slices input
            tags = None
            from_nodes = None
            to_nodes = None
            node_names = None
            from_inputs = None
            to_outputs = None
            node_namespace = None

            if slices:            
                for slice_item in slices:
                    slice_type = slice_item['slice']
                    slice_args = slice_item['args']
                    
                    if slice_type == 'tags':
                        tags = slice_args
                    if slice_type == 'from_nodes':
                        from_nodes = slice_args
                    if slice_type == 'to_nodes':
                        to_nodes = slice_args
                    if slice_type == 'node_names':
                        node_names = slice_args
                    if slice_type == 'from_inputs':
                        from_inputs = slice_args
                    if slice_type == 'to_outputs':
                        to_outputs = slice_args
                    if slice_type == 'node_namespace':
                        node_namespace = slice_args[0]

            record_data = {
                    "session_id": session.session_id,
                    "celery_task_id": self.request.id,
                    "project_path": session._project_path.as_posix(),
                    "env": session.load_context().env,
                    "kedro_version": kedro_version,
                    "tags": tags, # Construct the pipeline using only nodes which have this tag attached.
                    "from_nodes": from_nodes, # A list of node names which should be used as a starting point.
                    "to_nodes": to_nodes, # A list of node names which should be used as an end point.
                    "node_names": node_names, # Run only nodes with specified names.
                    "from_inputs": from_inputs, # A list of dataset names which should be used as a starting point.
                    "to_outputs": to_outputs, # A list of dataset names which should be used as an end point.
                    "load_versions": "", # Specify a particular dataset version (timestamp) for loading
                    "extra_params": "", # Specify extra parameters that you want to pass to the context initialiser.
                    "pipeline_name": name,
                    "namespace": node_namespace, # Name of the node namespace to run.
                    "runner": getattr(runner, "__name__", str(runner)),
                }
        
            hook_manager.hook.before_pipeline_run(
                    run_params=record_data,
                    pipeline=pipelines.get(name, None),
                    catalog=io
                )
            
            runner = init_runner(runner = runner)

            # Filter the pipeline based on the slices and only_missing parameters
            if only_missing:
                # https://github.com/kedro-org/kedro/blob/06d5a6920cbb6b45c07dca6f77d14bb35e283b19/kedro/runner/runner.py#L154
                free_outputs = pipelines[name].outputs() - set(io.list())
                missing = {ds for ds in io.list() if not io.exists(ds)}
                to_build = free_outputs | missing
                filtered_pipeline = pipelines[name].only_nodes_with_outputs(*to_build) + pipelines[name].from_inputs(
                    *to_build
                )

                # We also need any missing datasets that are required to run the
                # `filtered_pipeline` pipeline, including any chains of missing datasets.
                unregistered_ds = pipelines[name].datasets() - set(io.list())
                output_to_unregistered = pipelines[name].only_nodes_with_outputs(*unregistered_ds)
                input_from_unregistered = filtered_pipeline.inputs() & unregistered_ds
                filtered_pipeline += output_to_unregistered.to_outputs(*input_from_unregistered)
            else:
                filtered_pipeline = pipelines[name].filter(
                    tags=tags,
                    from_nodes=from_nodes,
                    to_nodes=to_nodes,
                    node_names=node_names,
                    from_inputs=from_inputs,
                    to_outputs=to_outputs,
                    node_namespace=node_namespace,
                )

            p = self.db.read(id=id)
            p.status[-1].filtered_nodes = [node.name for node in filtered_pipeline.nodes]
            self.db.update(p)

            run_result = runner().run(filtered_pipeline, catalog = io, hook_manager=hook_manager, session_id=session.session_id)

            hook_manager.hook.after_pipeline_run(
                    run_params=record_data,
                    run_result=run_result,
                    pipeline=pipelines.get(name, None),
                    catalog=io
                )

            return "success"
        except Exception as e:
            logger.exception(f"Error running pipeline: {e}")
            hook_manager.hook.on_pipline_error(
                error = e,
                run_params=record_data,
                pipeline=pipelines.get(name, None),
                catalog=io
            )
            raise e



import json
from functools import reduce
from urllib.parse import urlparse


def merge(a, b, path=None):
    """
    Merges nested dictionaries recursively.  Merges b into a.
    """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a


def merge_dicts(dicts):
    return reduce(merge, dicts)


def parse_s3_filepath(config):
    """
    Parse the s3 bucket name and key from DataSet filepath field.
    """
    if config:
        try:
            dataset_dict = json.loads(config)
            filepath = dataset_dict.get("filepath")
            if not filepath:
                raise ValueError("Invalid dataset configuration. Must have 'filepath' key")
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON in config: {e}")
    else:
        raise ValueError("Invalid dataset configuration. Must have 'filepath' key")

    if not filepath.startswith("s3://"):
        raise ValueError("Invalid S3 path. Must start with 's3://'")

    parsed = urlparse(filepath)
    bucket_name = parsed.netloc
    s3_key = parsed.path.lstrip("/")

    if not bucket_name:
        raise ValueError("Invalid S3 path. Bucket name is missing.")
    if not s3_key:
        raise ValueError("Invalid S3 path. S3 key (object path) is missing.")

    return bucket_name, s3_key



import os
from typing import Any

from kedro.framework.hooks import hook_impl
from kedro.io import CatalogProtocol
from kedro.pipeline import Pipeline

from .config import config


class InvalidPipeline(Exception):
    """Custom exception for invalid pipeline."""

    def __init__(self, message="The pipeline is invalid"):
        self.message = message
        super().__init__(self.message)


class DataValidationHooks:
    """
    A Kedro hook class to validate the existence of input datasets before running a pipeline.
    """

    @hook_impl
    def before_pipeline_run(
        self, run_params: dict[str, Any], pipeline: Pipeline, catalog: CatalogProtocol
    ) -> None:
        """Hook to be invoked before a pipeline runs.

        Args:
            run_params: The params used to run the pipeline.
            pipeline: The ``Pipeline`` that will be run.
            catalog: An implemented instance of ``CatalogProtocol`` to be used during the run.
        """

        # Only check the free inputs (excluding parameters) for validation
        filtered_keys = [key for key in pipeline.inputs() if not key.startswith('params:') and key != 'parameters']

        # Check to see if all free inputs exist
        for input in filtered_keys:
            if not catalog.exists(input):
                raise InvalidPipeline(f"Input dataset {input} does not exist.")


class DataLoggingHooks:
    """
    A Kedro hook class to save pipeline logs to S3 bucket. 
    """

    def save_meta(self, run_params: dict[str, Any], catalog: CatalogProtocol):
        d = catalog.load("gql_meta")
        d["run_params"] = run_params
        catalog.save("gql_meta", d)

    def save_logs(self, catalog: CatalogProtocol, session_id: str, celery_task_id: str):
        log_dir = os.path.join(config["KEDRO_GRAPHQL_LOG_TMP_DIR"], celery_task_id)
        log_files = ["info.log", "errors.log"]

        for log_file in log_files:
            log_path = os.path.join(log_dir, log_file)
            if os.path.exists(log_path) and os.path.getsize(log_path) > 0:
                with open(log_path, "r") as file:
                    logs = file.read()
                d = catalog._get_dataset("gql_logs")
                d.save({f"logs/{session_id}/{log_file}": logs})

    @hook_impl
    def before_pipeline_run(self, run_params: dict[str, Any], pipeline: Pipeline, catalog: CatalogProtocol):
        # Clear previous logs before pipeline run
        log_dir = os.path.join(config["KEDRO_GRAPHQL_LOG_TMP_DIR"], run_params["celery_task_id"])
        log_files = ["info.log", "errors.log"]

        for log_file in log_files:
            log_path = os.path.join(log_dir, log_file)
            if os.path.exists(log_path):
                open(log_path, 'w').close()

        if config.get('KEDRO_GRAPHQL_LOG_PATH_PREFIX'):
            self.save_meta(run_params, catalog)

    @hook_impl
    def after_pipeline_run(
            self, run_params: dict[str, Any],
            run_result: dict[str, Any],
            pipeline: Pipeline, catalog: CatalogProtocol):
        if config.get('KEDRO_GRAPHQL_LOG_PATH_PREFIX'):
            self.save_logs(catalog, run_params["session_id"], run_params["celery_task_id"])

    @hook_impl
    def on_pipline_error(self, error: Exception, run_params: dict[str, Any], pipeline: Pipeline, catalog: CatalogProtocol):
        if config.get('KEDRO_GRAPHQL_LOG_PATH_PREFIX'):
            self.save_logs(catalog, run_params["session_id"], run_params["celery_task_id"])



import celery


def celery_app(config, backend):
    app = celery.Celery()

    class Config:
        broker_url = config["KEDRO_GRAPHQL_BROKER"]
        result_backend = config["KEDRO_GRAPHQL_CELERY_RESULT_BACKEND"]
        result_extended = True
        task_serializer = 'json'
        result_serializer = 'json'
        accept_content = ['json']
        enable_utc = True
        worker_send_task_events = True
        task_send_sent_event = True
        imports = "kedro_graphql.tasks"

    app.config_from_object(Config)
    app.kedro_graphql_backend = backend
    return app


@celery.signals.setup_logging.connect
def on_setup_logging(**kwargs):
    #  Disable Celery logging configuration so logs are captured in info.log and error.log
    # https://docs.celeryq.dev/en/latest/userguide/tasks.html#logging
    pass



import json
import uuid
from copy import deepcopy
from datetime import datetime
from enum import Enum
from typing import List, Optional

import boto3
import strawberry
from botocore.exceptions import ClientError
from bson.objectid import ObjectId
from fastapi.encoders import jsonable_encoder
from kedro.io import AbstractDataset
from strawberry.scalars import JSON

from .config import config as CONFIG
from .utils import parse_s3_filepath


def mark_deprecated(default=None):
    return strawberry.field(default=default, deprecation_reason="see " + str(CONFIG["KEDRO_GRAPHQL_DEPRECATIONS_DOCS"]))


@strawberry.type
class Tag:
    key: str
    value: str


@strawberry.input
class TagInput:
    key: str
    value: str


@strawberry.enum
class ParameterType(Enum):
    STRING = "string"
    BOOLEAN = "boolean"
    INTEGER = "integer"
    FLOAT = "float"


@strawberry.type
class Parameter:
    name: str
    value: str
    type: Optional[ParameterType] = ParameterType.STRING

    @staticmethod
    def decode(input_dict) -> dict:
        """
        Returns a Parameter object from a dictionary.
        """
        if input_dict.get("type", True):
            return Parameter(
                name=input_dict["name"],
                value=input_dict["value"],
                type=ParameterType[input_dict["type"].upper()])
        else:
            return Parameter(**input_dict)

    def serialize(self) -> dict:
        """
        Returns serializable dict in format compatible with kedro.
        """
        value = self.value
        if self.type == "boolean":
            value = self.value.lower()
            if value == "true":
                value = True
            elif value == "false":
                value = False
            else:
                raise ValueError("Parameter of type BOOL must be one of 'True', 'true', 'False', or 'false'")

        elif self.type == "integer":
            value = int(self.value)

        elif self.type == "float":
            value = float(self.value)

        return {self.name: value}


@strawberry.input
class ParameterInput:
    name: str
    value: str
    type: Optional[ParameterType] = ParameterType.STRING

    @staticmethod
    def create(parameters: dict):
        params = []
        for k, v in parameters.items():
            ptype = type(v)

            if isinstance(ptype, int):
                ptype = ParameterType.INTEGER
            elif isinstance(ptype, float):
                ptype = ParameterType.FLOAT
            elif isinstance(ptype, bool):
                ptype = ParameterType.BOOLEAN
            else:
                ptype = ParameterType.STRING

            params.append(ParameterInput(name=k, value=str(v), type=ptype))
        return params


@strawberry.input
class CredentialSetInput:
    name: str
    value: str

    def serialize(self) -> dict:
        """
        Returns serializable dict in format compatible with kedro.
        """
        return {self.name: self.value}


@strawberry.input
class CredentialInput:
    name: str
    value: List[CredentialSetInput]

    def serialize(self) -> dict:
        """
        Returns serializable dict in format compatible with kedro.
        """
        values = {}
        for v in self.value:
            values.update(v.serialize())
        return {self.name: values}


@strawberry.input
class CredentialNestedInput:
    name: str
    value: List[CredentialInput]

    def serialize(self) -> dict:
        """
        Returns serializable dict in format compatible with kedro.
        """
        values = {}
        for v in self.value:
            values.update(v.serialize())
        return {self.name: values}


@strawberry.type
class DataSet:
    name: str
    config: Optional[str] = None
    tags: Optional[List[Tag]] = None

    @strawberry.field
    def pre_signed_url_create(self) -> Optional[JSON]:
        """
        Generate a presigned URL S3 to upload a file.

        Returns:
            Optional[JSON]: Dictionary with the URL to post to and form fields and values to submit with the POST. If an error occurs, returns None.

            Example:
                {
                    "url": "https://your-bucket-name.s3.amazonaws.com/",
                    "fields": {
                        "key": "your-object-key",
                        "AWSAccessKeyId": "your-access-key-id",
                        "x-amz-security-token": "your-security-token",
                        "policy": "your-policy",
                        "signature": "your-signature"
                    }
                }
        """

        bucket_name, s3_key = parse_s3_filepath(self.config)

        try:
            s3_client = boto3.client('s3')
            response = s3_client.generate_presigned_post(bucket_name,
                                                         f"{uuid.uuid4()}/{s3_key}",
                                                         Fields=None,
                                                         Conditions=None,
                                                         ExpiresIn=3600)
        except ClientError as e:
            print(f"Failed to generate presigned URL: {e}")
            return None

        return response

    @strawberry.field
    def pre_signed_url_read(self) -> Optional[str]:
        """
        Generate a presigned URL S3 to download a file.

        Returns:
            Optional[str]: download url with query parameters

            Example: https://your-bucket-name.s3.amazonaws.com/your-object-key?AWSAccessKeyId=your-access-key-id&Signature=your-signature&x-amz-security-token=your-security-token&Expires=expiration-time
        """

        bucket_name, s3_key = parse_s3_filepath(self.config)

        try:
            s3_client = boto3.client('s3')
            params = {
                'Bucket': bucket_name,
                'Key': s3_key
            }

            response = s3_client.generate_presigned_url(
                'get_object',
                Params=params,
                ExpiresIn=3600
            )
        except ClientError as e:
            print(f"Failed to generate presigned URL: {e}")
            return None

        return response

    @strawberry.field
    def exists(self) -> bool:
        if self.config:
            return AbstractDataset.from_config(self.name, json.loads(self.config)).exists()
        else:
            return False

    def serialize(self) -> dict:
        """
        Returns serializable dict in format compatible with kedro.
        """
        temp = self.__dict__.copy()
        temp.pop("name")
        return {self.name: json.loads(temp['config'])}

    @staticmethod
    def decode(payload):
        """
        Return a new DataSet from a dictionary.

        Args:
            payload (dict): dict representing DataSet e.g.

                {
                  "name": "text_in",
                  "config": '{"filepath": "./data/01_raw/text_in.txt", "type": "text.TextDataSet", "save_args": [{"name": "say", "value": "hello"}], "load_args": [{"name": "say", "value": "hello"}]}',
                  "tags":[{"key": "owner name", "value": "harinlee0803"},{"key": "owner email", "value": "test@example.com"}]
                }

        """
        if payload.get("tags", False):
            tags = [Tag(**t) for t in payload["tags"]]
        else:
            tags = None

        return DataSet(
            name=payload["name"],
            config=payload["config"],
            tags=tags
        )


@strawberry.input
class DataSetInput:
    name: str
    config: Optional[str] = None
    tags: Optional[List[TagInput]] = None


class DataCatalog:
    datasets: List[DataSet]


@strawberry.input
class DataCatalogInput:
    datasets: List[DataSetInput]

    @staticmethod
    def create(config):
        """
        context.config_loader["catalog"]

        {'text_in': {'type': 'text.TextDataSet',
                     'filepath': './data/01_raw/text_in.txt'},
         'text_out': {'type': 'text.TextDataSet',
                      'filepath': './data/02_intermediate/text_out.txt'}}

        Example usage:

            from kedro_graphql.models import DataCatalogInput

            catalog = DataCatalogInput.create(context.config_loader["catalog"])

            print(catalog)

            [DataSetInput(name='text_in', config='{"type": "text.TextDataSet", "filepath": "./data/01_raw/text_in.txt"}', type=None, filepath=None, save_args=None, load_args=None, credentials=None), 
             DataSetInput(name='text_out', config='{"type": "text.TextDataSet", "filepath": "./data/02_intermediate/text_out.txt"}', type=None, filepath=None, save_args=None, load_args=None, credentials=None)]

        """
        return [DataSetInput(name=k, config=json.dumps(v)) for k, v in config.items()]


@strawberry.type
class Node:
    name: str
    inputs: List[str]
    outputs: List[str]
    tags: List[str]


@strawberry.type(description="PipelineTemplates are definitions of Pipelines.  They represent the supported interface for executing a Pipeline.")
class PipelineTemplate:
    id: str = strawberry.field(description="ID of the pipeline template.")
    name: str
    kedro_pipelines: strawberry.Private[dict]
    kedro_catalog: strawberry.Private[dict]
    kedro_parameters: strawberry.Private[dict]

    @strawberry.field
    def describe(self) -> str:
        return self.kedro_pipelines[self.name].describe()

    @strawberry.field
    def nodes(self) -> List[Node]:
        nodes = self.kedro_pipelines[self.name].nodes

        return [Node(name=n.name, inputs=n.inputs, outputs=n.outputs, tags=n.tags) for n in nodes]

    @strawberry.field
    def parameters(self) -> List[Parameter]:
        # keep track of parameters to avoid returning duplicates
        params = {}
        for n in self.kedro_pipelines[self.name].all_inputs():
            if n.startswith("params:"):
                name = n.split("params:")[1]
                value = self.kedro_parameters[name]
                if not params.get(name, False):
                    params[name] = value
            elif n == "parameters":
                for k, v in self.kedro_parameters.items():
                    if not params.get(k, False):
                        params[k] = v
        return [Parameter(name=k, value=v) for k, v in params.items()]

    @strawberry.field
    def inputs(self) -> List[DataSet]:
        inputs_resolved = []
        for n in self.kedro_pipelines[self.name].all_inputs():
            if not n.startswith("params:") and n != "parameters":
                config = self.kedro_catalog[n]
                inputs_resolved.append(DataSet(name=n, config=json.dumps(config)))

        return inputs_resolved

    @strawberry.field
    def outputs(self) -> List[DataSet]:
        outputs_resolved = []
        for n in self.kedro_pipelines[self.name].all_outputs():
            config = self.kedro_catalog[n]
            outputs_resolved.append(DataSet(name=n, config=json.dumps(config)))

        return outputs_resolved


@strawberry.type
class PageMeta:
    next_cursor: Optional[str] = strawberry.field(
        description="The next cursor to continue with."
    )


@strawberry.type
class PipelineTemplates:
    pipeline_templates: List[PipelineTemplate] = strawberry.field(description="The list of pipeline templates.")

    page_meta: PageMeta = strawberry.field(description="Metadata to aid in pagination.")

    @staticmethod
    def _build_pipeline_index(kedro_pipelines, kedro_catalog, kedro_parameters):
        """
        """
        pipes = []
        count = 100000000000000000000000
        for k, v in kedro_pipelines.items():
            pipes.append(PipelineTemplate(name=k,
                                          id=ObjectId(str(count)),
                                          kedro_pipelines=kedro_pipelines,
                                          kedro_catalog=kedro_catalog,
                                          kedro_parameters=kedro_parameters))
            count += 1

        return pipes


@strawberry.enum
class PipelineSliceType(Enum):
    TAGS = "tags"
    FROM_NODES = "from_nodes"
    TO_NODES = "to_nodes"
    NODE_NAMES = "node_names"
    FROM_INPUTS = "from_inputs"
    TO_OUTPUTS = "to_outputs"
    NODE_NAMESPACE = "node_namespace"


@strawberry.input(description="Slice a pipeline.")
class PipelineSlice:
    slice: PipelineSliceType
    args: List[str]  # e.g. ["node1", "node2"]


@strawberry.enum
class PipelineInputStatus(Enum):
    STAGED = "STAGED"
    READY = "READY"


@strawberry.input(description="PipelineInput")
class PipelineInput:
    name: str
    state: PipelineInputStatus = PipelineInputStatus.STAGED
    parameters: Optional[List[ParameterInput]] = None
    data_catalog: Optional[List[DataSetInput]] = None
    tags: Optional[List[TagInput]] = None
    parent: Optional[uuid.UUID] = None
    runner: Optional[str] = None
    slices: Optional[List[PipelineSlice]] = None
    only_missing: Optional[bool] = False

    @staticmethod
    def create(name=None, data_catalog=None, parameters=None, tags=None):
        """
        Example usage:

            from kedro_graphql.models import PipelineInput
            from fastapi.encoders import jsonable_encoder

            p = PipelineInput(name = "example00",
                         data_catalog = context.config_loader["catalog"],
                         parameters = context.config_loader["parameters"],
                         tags = [{""owner":"person"}])

            print(p)

            PipelineInput(name='example00', 
                          parameters=[
                            ParameterInput(name='example', 
                                           value='hello', 
                                           type=<ParameterType.STRING: 'string'>), 
                            ParameterInput(name='duration', value='1', type=<ParameterType.STRING: 'string'>)
                          ], 
                          data_catalog=[
                            DataSetInput(name='text_in', config='{"type": "text.TextDataSet", "filepath": "./data/01_raw/text_in.txt"}'), 
                            DataSetInput(name='text_out', config='{"type": "text.TextDataSet", "filepath": "./data/02_intermediate/text_out.txt"}')
                          ], 
                          tags=[TagInput(key='owner', value='sean')])

            print(jsonable_encoder(p))

            ## this can be used as the PipelineInput parameter when calleing the pipeline mutation via the API
            {'name': 'example00',
            'parameters': [{'name': 'example', 'value': 'hello', 'type': 'string'},
             {'name': 'duration', 'value': '1', 'type': 'string'}],
            'data_catalog': [{'name': 'text_in',
              'config': '{"type": "text.TextDataSet", "filepath": "./data/01_raw/text_in.txt"}'},
             {'name': 'text_out',
              'config': '{"type": "text.TextDataSet", "filepath": "./data/02_intermediate/text_out.txt"}'}],
            'tags': [{'key': 'owner', 'value': 'sean'}],
            'credentials': None,
            'credentials_nested': None}

        """
        if tags:
            tags = [TagInput(key=k, value=v) for t in tags for k, v in t.items()]

        if data_catalog:
            data_catalog = DataCatalogInput.create(data_catalog)

        if parameters:
            parameters = ParameterInput.create(parameters)

        return PipelineInput(name=name,
                             parameters=parameters,
                             data_catalog=data_catalog,
                             tags=tags)


@strawberry.enum
class State(Enum):
    READY = 'READY'
    STAGED = 'STAGED'
    STARTED = 'STARTED'
    RETRY = 'RETRY'
    FAILURE = 'FAILURE'
    SUCCESS = 'SUCCESS'
    REVOKED = 'REVOKED'
    PENDING = 'PENDING'
    RECIEVED = 'RECIEVED'


@strawberry.type
class PipelineStatus:
    state: State
    session: Optional[str]
    runner: Optional[str] = None
    filtered_nodes: Optional[List[str]] = None
    started_at: Optional[datetime] = None
    finished_at: Optional[datetime] = None
    task_id: Optional[str] = None
    task_name: Optional[str] = None
    task_args: Optional[str] = None
    task_kwargs: Optional[str] = None
    task_request: Optional[str] = None
    task_exception: Optional[str] = None
    task_traceback: Optional[str] = None
    task_einfo: Optional[str] = None
    task_result: Optional[str] = None


@strawberry.type
class Pipeline:
    kedro_pipelines_index: strawberry.Private[Optional[List[PipelineTemplate]]] = None
    id: Optional[uuid.UUID] = None
    name: str
    data_catalog: Optional[List[DataSet]] = None
    parameters: Optional[List[Parameter]] = None
    status: List[PipelineStatus] = strawberry.field(default_factory=list)
    tags: Optional[List[Tag]] = None
    created_at: Optional[datetime] = None
    parent: Optional[uuid.UUID] = None
    project_version: Optional[str] = None
    pipeline_version: Optional[str] = None
    kedro_graphql_version: Optional[str] = None

    @strawberry.field
    def template(self) -> PipelineTemplate:
        for p in self.kedro_pipelines_index:
            if p.name == self.name:
                return p

    @strawberry.field
    def describe(self) -> str:
        return self.template().describe()

    @strawberry.field
    def nodes(self) -> List[Node]:
        return self.template().nodes()

    def serialize(self):
        parameters = {}
        data_catalog = {}

        if self.parameters:
            for p in self.parameters:
                s = p.serialize()
                parameters.update(s)

        if self.data_catalog:
            for d in self.data_catalog:
                s = d.serialize()
                data_catalog.update(s)

        return {
            "id": self.id,
            "name": self.name,
            "data_catalog": data_catalog,
            "parameters": parameters,
        }

    def encode(self, encoder="dict"):

        if encoder == "dict":
            p = deepcopy(self)
            p.id = str(p.id)  # if type ObjectID the jsonable_encoder will throw an error
            encoded_pipeline = jsonable_encoder(p)

            # remove fields we dont want to encode
            encoded_pipeline.pop("kedro_pipelines_index")
            return encoded_pipeline
        elif encoder == "kedro":
            return self.serialize()
        else:
            raise TypeError("encoder must be 'dict' or 'kedro'")

    @staticmethod
    def decode(payload):
        if payload.get("tags", None):
            tags = [Tag(**t) for t in payload["tags"]]
        else:
            tags = None

        if payload.get("data_catalog", None):
            data_catalog = [DataSet.decode(d) for d in payload["data_catalog"]]
        else:
            data_catalog = []

        if payload.get("status", None):
            status = [PipelineStatus(
                state=State[s["state"]],
                session=s["session"],
                runner=s.get("runner", "kedro.runner.SequentialRunner"),
                filtered_nodes=s.get("filtered_nodes"),
                started_at=datetime.fromisoformat(s["started_at"]) if s.get("started_at") else None,
                finished_at=datetime.fromisoformat(s["finished_at"]) if s.get("finished_at") else None,
                task_name=s.get("task_name"),
                task_id=s.get("task_id"),
                task_args=s.get("task_args"),
                task_kwargs=s.get("task_kwargs"),
                task_request=s.get("task_request"),
                task_exception=s.get("task_exception"),
                task_traceback=s.get("task_traceback"),
                task_einfo=s.get("task_einfo"),
                task_result=s.get("task_result")
            ) for s in payload["status"]]
        else:
            status = []

        if payload.get("parameters", None):
            parameters = [Parameter.decode(p) for p in payload["parameters"]]
        else:
            parameters = None

        return Pipeline(
            id=payload.get("id", None),
            name=payload["name"],
            data_catalog=data_catalog,
            parameters=parameters,
            status=status,
            tags=tags,
            created_at=datetime.fromisoformat(payload["created_at"]) if payload.get("created_at", None) else None,
            parent=payload.get("parent", None),
            project_version=payload.get("project_version", None),
            pipeline_version=payload.get("pipeline_version", None),
            kedro_graphql_version=payload.get("kedro_graphql_version", None)
        )


@strawberry.type
class Pipelines:
    pipelines: List[Pipeline] = strawberry.field(description="The list of pipeline instances.")
    page_meta: PageMeta = strawberry.field(description="Metadata to aid in pagination.")


@strawberry.type
class PipelineEvent:
    id: str
    task_id: str
    status: str
    result: Optional[str] = None
    timestamp: str
    traceback: Optional[str] = None


@strawberry.type
class PipelineLogMessage:
    id: str
    message: str
    message_id: str
    task_id: str
    time: str



from setuptools import find_packages, setup


setup(
    name="kedro_graphql",
    packages=find_packages(exclude=["tests"]),
    extras_require={
        "docs": [
            "docutils<0.18.0",
            "sphinx~=3.4.3",
            "sphinx_rtd_theme==0.5.1",
            "nbsphinx==0.8.1",
            "nbstripout~=0.4",
            "sphinx-autodoc-typehints==1.11.1",
            "sphinx_copybutton==0.3.1",
            "ipykernel>=5.3, <7.0",
            "Jinja2<3.1.0",
            "myst-parser~=0.17.2",
        ]
    },
)



import abc
import uuid

from kedro_graphql.models import Pipeline


class BaseBackend(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def startup(self, **kwargs):
        """Startup hook."""
        raise NotImplementedError

    @abc.abstractmethod
    def shutdown(self, **kwargs):
        """Shutdown hook."""
        raise NotImplementedError

    @abc.abstractmethod
    def read(self, id: uuid.UUID = None, task_id: str = None):
        """Load a pipeline by id"""
        raise NotImplementedError

    @abc.abstractmethod
    def list(self, cursor: uuid.UUID = None, limit: int = None, filter: str = None, sort: str = None):
        """List pipelines using cursor pagination"""
        raise NotImplementedError

    @abc.abstractmethod
    def create(self, pipeline: Pipeline):
        """Save a pipeline"""
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, pipeline: Pipeline):
        """Update a pipeline"""
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, id: uuid.UUID = None):
        """Delete a pipeline"""
        raise NotImplementedError



from kedro_graphql.models import State


def test_backend_create(mock_app, mock_pipeline_no_task):
    p = mock_app.backend.create(mock_pipeline_no_task)
    assert p.id is not None
    p.id = None
    assert p == mock_pipeline_no_task


def test_backend_update(mock_app, mock_pipeline_no_task):
    p = mock_app.backend.create(mock_pipeline_no_task)
    p.name = "example01"
    p = mock_app.backend.update(p)
    assert p.name == "example01"


def test_backend_update_status(mock_app, mock_pipeline_no_task):
    p = mock_app.backend.create(mock_pipeline_no_task)
    p.status[-1].state = State.STARTED
    p = mock_app.backend.update(p)
    assert p.status[-1].state == State.STARTED



import asyncio
import logging
import time
from queue import Empty as QueueEmptyException
from queue import Queue
from threading import Thread
from typing import AsyncGenerator

from celery.states import READY_STATES

logger = logging.getLogger("kedro-graphql")


class PipelineEventMonitor:
    def __init__(self, app=None, task_id=None, timeout=1):
        """
        Kwargs:
            app (Celery): celery application instance.
            uuid (str): a celery task id.
            timeout (float): See https://docs.python.org/3/library/queue.html#queue.Queue.get
        """
        self.task_id = task_id
        self.app = app
        self.timeout = timeout

    @staticmethod
    def _task_event_receiver(app, queue, task_id):
        """
        Recieves task events from backend broker and puts them in a 
        Queue.  Incoming tasks are filtered and only tasks with a 
        root_id or uuid matching the provided id are put in the Queue.

        Example event payloads:

        {'hostname': 'gen36975@alligator', 'utcoffset': 5, 'pid': 36975, 'clock': 7864, 'uuid': 'd8253d45-ce28-4719-b2ba-8e266dfdaf04', 'root_id': 'd8253d45-ce28-4719-b2ba-8e266dfdaf04', 'parent_id': None, 'name': 'kedro_graphql.tasks.run_pipeline', 'args': '()', 'kwargs': "{'name': 'example00', 'inputs': {'text_in': {'type': 'text.TextDataSet', 'filepath': './data/01_raw/text_in.txt'}}, 'outputs': {'text_out': {'type': 'text.TextDataSet', 'filepath': './data/02_intermediate/text_out.txt'}}}", 'retries': 0, 'eta': None, 'expires': None, 'queue': 'celery', 'exchange': '', 'routing_key': 'celery', 'timestamp': 1672860581.1371481, 'type': 'task-sent', 'local_received': 1672860581.138474}
        {'hostname': 'celery@alligator', 'utcoffset': 5, 'pid': 37029, 'clock': 7867, 'uuid': 'd8253d45-ce28-4719-b2ba-8e266dfdaf04', 'timestamp': 1672860581.1411166, 'type': 'task-started', 'local_received': 1672860581.144976}
        {'hostname': 'celery@alligator', 'utcoffset': 5, 'pid': 37029, 'clock': 7870, 'uuid': 'd8253d45-ce28-4719-b2ba-8e266dfdaf04', 'result': "'success'", 'runtime': 2.013245126003312, 'timestamp': 1672860583.1549191, 'type': 'task-succeeded', 'local_received': 1672860583.158338}

        Args:
            app (Celery): celery application instance.
            queue (Queue):  a python queue.Queue.
            task_id (str):  celery task id.

        """

        def process_tasks(event):
            if event.get("root_id", "") == task_id or event.get("uuid") == task_id:
                queue.put(event)

        with app.connection() as connection:
            recv = app.events.Receiver(connection, handlers={
                'task-sent': process_tasks,
                'task-recieved': process_tasks,
                'task-started': process_tasks,
                'task-succeeded': process_tasks,
                'task-failed': process_tasks,
                'task-rejected': process_tasks,
                'task-revoked': process_tasks,
                'task-retried': process_tasks
            })

            recv.capture(limit=None, timeout=None, wakeup=True)

    def _start_task_event_receiver_thread(self, queue):
        """
        Start the task event receiver in a thread.

        Args:
            queue (Queue): a python queue.Queue.

        Returns:
            worker (threading.Thread): a python thread object.

        """
        worker = Thread(target=self._task_event_receiver, args=(self.app, queue, self.task_id))
        worker.daemon = True
        worker.start()
        logger.info("started event reciever thread")
        return worker

    async def consume(self) -> AsyncGenerator[dict, None]:
        q = Queue()

        event_thread = self._start_task_event_receiver_thread(q)
        # https://docs.celeryq.dev/en/stable/reference/celery.events.state.html#module-celery.events.state
        state = self.app.events.State()

        while True:
            try:
                event = q.get(timeout=self.timeout)
                state.event(event)
                # task name is sent only with -received event, and state
                # will keep track of this for us.
                task = state.tasks.get(event['uuid'])
                yield {"task_id": task.id, "status": task.state, "result": task.result, "timestamp": task.timestamp, "traceback": task.traceback}
                q.task_done()
                if task.state in READY_STATES:
                    break
            except QueueEmptyException:
                if self.app.AsyncResult(self.task_id).status in READY_STATES:
                    break
                else:
                    continue

        event_thread.join(timeout=0.1)

    async def start(self, interval=0.5) -> AsyncGenerator[dict, None]:
        """
        A simplified but fully async version of the PipelineEventMonitor().consume() method.

        The PipelineEventMonitor.consume() method relies on celery's native
        real time event processing approach which is syncronous and blocking.
        https://docs.celeryq.dev/en/stable/userguide/monitoring.html#real-time-processing

        """

        while True:
            task = self.app.AsyncResult(self.task_id)
            yield {"task_id": task.id, "status": task.state, "result": task.result, "timestamp": time.time(), "traceback": task.traceback}
            if self.app.AsyncResult(self.task_id).status in READY_STATES:
                break
            await asyncio.sleep(interval)



import json
import logging
import os
from inspect import currentframe, getframeinfo
from logging import LogRecord
from typing import AsyncGenerator

import redis
import redis.asyncio as redis_asyncio
from celery.result import AsyncResult
from celery.states import READY_STATES

from .json_log_formatter import JSONFormatter  # VerboseJSONFormatter also available

logger = logging.getLogger("kedro-graphql")
logger.setLevel(logging.INFO)


class RedisLogStreamPublisher(object):
    def __init__(self, topic, broker_url=None):
        self.connection = redis.Redis.from_url(broker_url)
        self.topic = topic
        if not self.connection.exists(self.topic):
            self.connection.xadd(
                self.topic, json.loads(
                    JSONFormatter().format(
                        LogRecord(
                            "kedro-graphql", 20, os.path.abspath(__file__),
                            getframeinfo(currentframe()).lineno, "Starting log stream", None, None))))
            # stream will expire in 24 hours (safety mechanism in case task_postrun fails to delete)
            self.connection.expire(self.topic, 86400)

    def publish(self, data):
        data = {k: (str(v) if isinstance(v, bool) else v) for k, v in data.items()}
        self.connection.xadd(self.topic, data)


class RedisLogStreamSubscriber(object):

    @classmethod
    async def create(cls, topic, broker_url=None):
        """Factory method for async instantiation RedisLogStreamSubscriber objects.
        """
        self = RedisLogStreamSubscriber()
        self.topic = topic
        self.broker_url = broker_url
        self.connection = await redis_asyncio.from_url(broker_url)
        return self

    async def consume(self, count=1, start_id=0):
        r = await self.connection.xread(count=count, streams={self.topic: start_id})
        return r


class KedroGraphQLLogHandler(logging.StreamHandler):
    def __init__(self, topic, broker_url=None):
        logging.StreamHandler.__init__(self)
        self.broker_url = broker_url
        self.topic = topic
        self.broker = RedisLogStreamPublisher(topic, broker_url=broker_url)
        self.setFormatter(JSONFormatter())

    def emit(self, record):
        msg = self.format(record)
        self.broker.publish(json.loads(msg))


class PipelineLogStream():

    @classmethod
    async def create(cls, task_id, broker_url=None):
        """Factory method for async instantiation PipelineLogStream objects.
        """
        self = PipelineLogStream()
        self.task_id = task_id
        self.broker_url = broker_url
        self.broker = await RedisLogStreamSubscriber().create(task_id, broker_url)
        return self

    async def consume(self) -> AsyncGenerator[dict, None]:
        start_id = 0
        while True:
            stream_data = await self.broker.consume(count=1, start_id=start_id)
            if len(stream_data) > 0:
                for id, value in stream_data[0][1]:
                    yield {"task_id": self.task_id, "message_id": id.decode(), "message": value[b"message"].decode(), "time": value[b"time"].decode()}
                # https://redis-py.readthedocs.io/en/stable/examples/redis-stream-example.html#read-more-data-from-the-stream
                start_id = stream_data[0][1][-1][0]
            else:
                # check task status
                r = AsyncResult(self.task_id)
                if r.status in READY_STATES:
                    await self.broker.connection.close()
                    break



from fastapi import FastAPI
from strawberry.fastapi import GraphQLRouter

from .backends import init_backend
from .celeryapp import celery_app
from .config import config
from .decorators import RESOLVER_PLUGINS, TYPE_PLUGINS, discover_plugins
from .models import PipelineTemplates
from .schema import build_schema


class KedroGraphQL(FastAPI):

    def __init__(self, kedro_session=None, config=config):
        super(KedroGraphQL, self).__init__()

        self.kedro_session = kedro_session
        self.kedro_context = self.kedro_session.load_context()
        self.kedro_catalog = self.kedro_context.config_loader["catalog"]
        self.kedro_parameters = self.kedro_context.config_loader["parameters"]
        from kedro.framework.project import pipelines
        self.kedro_pipelines = pipelines
        self.kedro_pipelines_index = PipelineTemplates._build_pipeline_index(
            self.kedro_pipelines, self.kedro_catalog, self.kedro_parameters)

        self.config = config

        self.resolver_plugins = RESOLVER_PLUGINS
        self.type_plugins = TYPE_PLUGINS

        discover_plugins(self.config)
        self.schema = build_schema(self.type_plugins)
        self.backend = init_backend(self.config)
        self.graphql_app = GraphQLRouter(self.schema)
        self.include_router(self.graphql_app, prefix="/graphql")
        self.add_websocket_route("/graphql", self.graphql_app)
        self.celery_app = celery_app(self.config, self.backend)

        @self.on_event("startup")
        def startup_backend():
            self.backend.startup()

        @self.on_event("shutdown")
        def shutdown_backend():
            self.backend.shutdown()



import json
import time

import pytest

IN_DEV = True


class TestSchemaMutations:
    create_pipeline_mutation = """
        mutation CreatePipeline($pipeline: PipelineInput!) {
          createPipeline(pipeline: $pipeline) {
              id
              name
              describe
              nodes {
                name
                inputs
                outputs
                tags
              }
              parameters {
                name
                value
                type
              }
              status {
                finishedAt
                state
                runner
                session
                startedAt
                taskArgs
                taskEinfo
                taskException
                taskId
                taskKwargs
                taskName
                taskRequest
                taskResult
                taskTraceback
                filteredNodes
              }
              tags {
                key
                value
              }
              parent
              createdAt
              dataCatalog {
                name
                config
              }
          }
        }
        """

    update_pipeline_mutation = """
        mutation UpdatePipeline($pipeline: PipelineInput!, $id: String!) {
          updatePipeline(pipeline: $pipeline, id: $id) {
            id
            name
            describe
            nodes {
              name
              inputs
              outputs
              tags
            }
            parameters {
              name
              value
              type
            }
            status {
              finishedAt
              state
              runner
              session
              startedAt
              taskArgs
              taskEinfo
              taskException
              taskId
              taskKwargs
              taskName
              taskRequest
              taskResult
              taskTraceback
              filteredNodes
            }
            tags {
              key
              value
            }
            parent
            createdAt
            dataCatalog {
                name
                config
            }
          }
        }
        """

    delete_pipeline_mutation = """
        mutation TestMutation($id: String!) {
          deletePipeline(id: $id) {
           id
            name
            describe
            nodes {
              name
              inputs
              outputs
              tags
            }
            parameters {
              name
              value
              type
            }
            status {
              finishedAt
              state
              runner
              session
              startedAt
              taskArgs
              taskEinfo
              taskException
              taskId
              taskKwargs
              taskName
              taskRequest
              taskResult
              taskTraceback
              filteredNodes
            }
            tags {
              key
              value
            }
            parent
            createdAt
            dataCatalog {
                name
                config
            }
          }
        }
        """

    @pytest.mark.usefixtures('mock_celery_session_app')
    @pytest.mark.usefixtures('celery_session_worker')
    @pytest.mark.usefixtures('depends_on_current_app')
    @pytest.mark.asyncio
    async def test_pipeline(self, mock_app, mock_info_context, mock_text_in, mock_text_out):

        resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                             variable_values={"pipeline": {
                                                 "name": "example00",
                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                 {"name": "text_out", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                 ],
                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                             }})

        assert resp.errors is None

    @pytest.mark.skipif(IN_DEV, reason="credential support in development")
    @pytest.mark.asyncio
    async def test_pipeline_creds(self, mock_app, mock_info_context, mock_text_in, mock_text_out):

        resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                             variable_values={"pipeline": {
                                                 "name": "example00",
                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in), "credentials": "my_creds"})},
                                                                 {"name": "text_out", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                 ],
                                                 "parameters": [{"name": "example", "value": "hello"}],
                                                 "tags": [{"key": "author", "value": "opensean"},
                                                          {"key": "package", "value": "kedro-graphql"}],
                                                 "credentials": [{"name": "my_creds",
                                                                  "value": [
                                                                      {"name": "key", "value": "admin"},
                                                                      {"name": "secret", "value": "password"}
                                                                  ]}],
                                                 "credentialsNested": [{"name": "my_creds",
                                                                        "value": [{"name": "client_kwargs",
                                                                                   "value": [{"name": "endpoint_url",
                                                                                              "value": "http://localhost:9000"
                                                                                              }]
                                                                                   }]
                                                                        }]
                                             }})

        assert resp.errors is None

    @pytest.mark.usefixtures('mock_celery_session_app')
    @pytest.mark.usefixtures('celery_session_worker')
    @pytest.mark.usefixtures('depends_on_current_app')
    @pytest.mark.asyncio
    async def test_pipeline2(self, mock_app, mock_info_context, mock_text_in_tsv, mock_text_out_tsv):

        resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                             variable_values={"pipeline": {
                                                 "name": "example00",
                                                 "dataCatalog": [
                                                     {"name": "text_in", "config": json.dumps({"type": "pandas.CSVDataset", "filepath": str(mock_text_in_tsv), "loadArgs": [{"name": "sep", "value": "\t"}], "saveArgs": [{"name": "sep", "value": "\t"}]})},
                                                     {"name": "text_out", "config": json.dumps({"type": "pandas.CSVDataset", "filepath": str(mock_text_out_tsv), "loadArgs": [{"name": "sep", "value": "\t"}], "saveArgs": [{"name": "sep", "value": "\t"}]})}
                                                 ],
                                                 "parameters": [{"name": "example", "value": "hello"}],
                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                             }})

        assert resp.errors is None

    @pytest.mark.usefixtures('mock_celery_session_app')
    @pytest.mark.usefixtures('celery_session_worker')
    @pytest.mark.usefixtures('depends_on_current_app')
    @pytest.mark.asyncio
    async def test_create_staged_pipeline(self, mock_app, mock_info_context, mock_text_in, mock_text_out):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example00",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "text_out", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                                 ],
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "state": "STAGED",
                                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                                             }})

        pipeline_state = create_pipeline_resp.data["createPipeline"]["status"][-1]["state"]

        assert create_pipeline_resp.errors is None
        assert pipeline_state == "STAGED"

    @pytest.mark.usefixtures('mock_celery_session_app')
    @pytest.mark.usefixtures('celery_session_worker')
    @pytest.mark.usefixtures('depends_on_current_app')
    @pytest.mark.asyncio
    async def test_create_valid_ready_pipeline(self, mock_app, mock_info_context, mock_text_in, mock_text_out):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example00",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "text_out", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                                 ],
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "state": "READY",
                                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                                             }})

        pipeline_state = create_pipeline_resp.data["createPipeline"]["status"][-1]["state"]

        assert create_pipeline_resp.errors is None
        assert pipeline_state != "STAGED"

    @pytest.mark.usefixtures('mock_celery_session_app')
    @pytest.mark.usefixtures('celery_session_worker')
    @pytest.mark.usefixtures('depends_on_current_app')
    @pytest.mark.asyncio
    async def test_create_invalid_name_ready_pipeline(self, mock_app, mock_info_context, mock_text_in, mock_text_out):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example02",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "text_out", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                                 ],
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "state": "READY",
                                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                                             }})

        assert create_pipeline_resp.errors is not None

    @pytest.mark.usefixtures('mock_celery_session_app')
    @pytest.mark.usefixtures('celery_session_worker')
    @pytest.mark.usefixtures('depends_on_current_app')
    @pytest.mark.asyncio
    async def test_update_pipeline_staged_to_ready(self, mock_app, mock_info_context, mock_text_in, mock_text_out):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example00",
                                                                 "state": "STAGED",
                                                             }})

        pipeline_id = create_pipeline_resp.data["createPipeline"]["id"]

        update_pipeline_resp = await mock_app.schema.execute(self.update_pipeline_mutation,
                                                             variable_values={"id": pipeline_id,
                                                                              "pipeline": {
                                                                                  "name": "example00",
                                                                                  "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                                  {"name": "text_out", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                                                  ],
                                                                                  "parameters": [{"name": "example", "value": "hello"},
                                                                                                 {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                                  "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}],
                                                                                  "state": "READY",
                                                                              }
                                                                              })
        assert update_pipeline_resp.errors is None
        assert update_pipeline_resp.data["updatePipeline"]["status"][-1]["state"] != "STAGED"
        assert update_pipeline_resp.data["updatePipeline"]["dataCatalog"][0] == {
            "name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})}
        assert update_pipeline_resp.data["updatePipeline"]["dataCatalog"][1] == {
            "name": "text_out", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_out)})}
        assert update_pipeline_resp.data["updatePipeline"]["parameters"][0] == {
            "name": "example", "value": "hello", "type": "STRING"}
        assert update_pipeline_resp.data["updatePipeline"]["parameters"][1] == {
            "name": "duration", "value": "0.1", "type": "FLOAT"}
        assert update_pipeline_resp.data["updatePipeline"]["tags"][0] == {"key": "author", "value": "opensean"}
        assert update_pipeline_resp.data["updatePipeline"]["tags"][1] == {"key": "package", "value": "kedro-graphql"}

    @pytest.mark.usefixtures('mock_celery_session_app')
    @pytest.mark.usefixtures('celery_session_worker')
    @pytest.mark.usefixtures('depends_on_current_app')
    @pytest.mark.asyncio
    async def test_delete_pipeline(self, mock_app, mock_info_context, mock_text_in, mock_text_out):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example00",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "text_out", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                                 ],
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                                             }})

        pipeline_id = create_pipeline_resp.data["createPipeline"]["id"]

        delete_pipeline_resp = await mock_app.schema.execute(self.delete_pipeline_mutation,
                                                             variable_values={"id": pipeline_id})

        assert delete_pipeline_resp.errors is None

    @pytest.mark.usefixtures('mock_celery_session_app')
    @pytest.mark.usefixtures('celery_session_worker')
    @pytest.mark.usefixtures('depends_on_current_app')
    @pytest.mark.asyncio
    async def test_pipeline_data_catalog_modified_with_log_datasets(
            self, mock_app, mock_info_context, mock_text_in, mock_text_out):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example00",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "text_out", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                                 ],
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "state": "READY",

                                                             }})
        # Sleep for 1 second to ensure pipeline before_start handler has run
        time.sleep(1)

        dataset_names = {ds.name for ds in mock_app.backend.read(
            create_pipeline_resp.data["createPipeline"]["id"]).data_catalog}

        assert "gql_meta" in dataset_names
        assert "gql_logs" in dataset_names

    @pytest.mark.usefixtures('mock_celery_session_app')
    @pytest.mark.usefixtures('celery_session_worker')
    @pytest.mark.usefixtures('depends_on_current_app')
    @pytest.mark.asyncio
    async def test_pipeline_slicing(
            self, mock_app, mock_info_context, mock_text_in, mock_uppercased_txt, mock_reversed_txt, mock_timestamped_txt):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example01",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "uppercased", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_uppercased_txt)})},
                                                                                 {"name": "reversed", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_reversed_txt)})},
                                                                                 {"name": "timestamped", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_timestamped_txt)})}],
                                                                 "slices": {"slice": "NODE_NAMES", "args": ["uppercase_node", "reverse_node"]},
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "state": "READY",
                                                             }})

        # Sleep for 1 second to ensure pipeline filtered_nodes has been updated
        time.sleep(1)

        # Make sure only nodes specified in "slices" were run
        assert mock_app.backend.read(create_pipeline_resp.data["createPipeline"]
                                     ["id"]).status[-1].filtered_nodes == ["uppercase_node", "reverse_node"]
        create_pipeline_resp.errors is None

    @pytest.mark.usefixtures('mock_celery_session_app')
    @pytest.mark.usefixtures('celery_session_worker')
    @pytest.mark.usefixtures('depends_on_current_app')
    @pytest.mark.asyncio
    async def test_pipeline_run_only_missing(
            self, mock_app, mock_info_context, mock_text_in, mock_uppercased_txt, mock_reversed_txt, mock_timestamped_txt):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example01",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "uppercased", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_uppercased_txt)})},
                                                                                 {"name": "reversed", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_reversed_txt)})},
                                                                                 {"name": "timestamped", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_timestamped_txt)})}],
                                                                 "onlyMissing": True,
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "state": "READY",
                                                             }})

        # Sleep for 1 second to ensure pipeline filtered_nodes has been updated
        time.sleep(1)

        # Make sure only timestamp_node was run because the file does not exist (did not write to it in conftest.py)
        assert mock_app.backend.read(create_pipeline_resp.data["createPipeline"]
                                     ["id"]).status[-1].filtered_nodes == ["timestamp_node"]
        create_pipeline_resp.errors is None



import pytest


@pytest.mark.usefixtures('mock_celery_session_app')
@pytest.mark.usefixtures('celery_session_worker')
@pytest.mark.usefixtures('depends_on_current_app')
class TestSchemaSubscriptions:
    @pytest.mark.asyncio
    async def test_pipeline(self, mock_app, mock_info_context, mock_pipeline):
        """Requires Redis to run.
        """

        query = """
    	  subscription {
          	pipeline(id:""" + '"' + str(mock_pipeline.id) + '"' + """) {
              id
              taskId
              status
              result
              timestamp
              traceback
            }
    	  }
        """
        sub = await mock_app.schema.subscribe(query)

        async for result in sub:
            assert not result.errors

    @pytest.mark.asyncio
    async def test_pipeline_logs(self, mock_app, mock_info_context, mock_pipeline, mock_pipeline2):
        """Requires Redis to run.
        This test runs two pipelines simultaneously to ensure logs messages are scoped
        to the correct pipeline.

        """

        query = """
    	  subscription {
          	pipelineLogs(id:""" + '"' + str(mock_pipeline.id) + '"' + """) {
              id
              message
              messageId
              taskId
              time
            }
    	  }
        """

        sub = await mock_app.schema.subscribe(query)

        async for result in sub:
            assert not result.errors
            assert result.data["pipelineLogs"]["id"] == str(mock_pipeline.id)
            assert result.data["pipelineLogs"]["taskId"] == str(mock_pipeline.status[-1].task_id)

        query2 = """
    	  subscription {
          	pipelineLogs(id:""" + '"' + str(mock_pipeline2.id) + '"' + """) {
              id
              message
              messageId
              taskId
              time
            }
    	  }
        """

        sub2 = await mock_app.schema.subscribe(query2)

        async for result in sub2:
            assert not result.errors
            assert result.data["pipelineLogs"]["id"] == str(mock_pipeline2.id)
            assert result.data["pipelineLogs"]["taskId"] == str(mock_pipeline2.status[-1].task_id)



"""
This is a boilerplate pipeline 'example01'
generated using Kedro 0.19.11
"""

from kedro.pipeline import Pipeline, node, pipeline

from .nodes import append_timestamp, reverse, uppercase


def create_pipeline(**kwargs) -> Pipeline:
    return pipeline([
        node(uppercase, inputs="text_in", outputs="uppercased", name="uppercase_node"),
        node(reverse, inputs="uppercased", outputs="reversed", name="reverse_node"),
        node(append_timestamp, inputs="reversed", outputs="timestamped", name="timestamp_node"),
    ])



import json

import pytest
from omegaconf import OmegaConf

from kedro_graphql.models import DataSet, Parameter


class TestDataSet:

    def test_serialize(self):
        params = {"name": "text_in",
                  "config": json.dumps({"type": "text.TextDataset",
                                        "filepath": "/tmp/test_in.csv",
                                        "load_args": {"delimiter": "\t"},
                                        "save_args": {"delimiter": "\t"}})
                  }

        expected = {"text_in": {
            "type": "text.TextDataset",
            "filepath": "/tmp/test_in.csv",
            "load_args": {
                "delimiter": "\t"
            },
            "save_args": {
                "delimiter": "\t"
            }
        }
        }

        d = DataSet(**params)
        output = d.serialize()

        assert output == expected

    def test_pre_signed_url_create_config_no_filepath(self):
        d = DataSet(name="text_in", config='{"type":"text.TextDataset"}')

        with pytest.raises(ValueError):
            output = d.pre_signed_url_create()

    def test_pre_signed_url_read_config_no_filepath(self):
        d = DataSet(name="text_in", config='{"type":"text.TextDataset"}')

        with pytest.raises(ValueError):
            output = d.pre_signed_url_read()

    def test_does_exist_with_config(self, mock_text_in):
        params = {
            "name": "text_in",
            "config": f'{{"type": "text.TextDataset", "filepath": "{str(mock_text_in)}"}}'
        }

        d = DataSet(**params)
        assert d.exists() == True

    def test_does_not_exist_with_config(self):
        params = {
            "name": "text_in",
            "config": '{"type": "text.TextDataset", "filepath": "/tmp/does_not_exist.csv"}'
        }

        d = DataSet(**params)
        assert d.exists() == False


class TestParameter:

    def test_serialize_string(self):
        params = {
            "name": "delimiter",
                    "value": "\t",
                    "type": "string"
        }

        expected = {
            "delimiter": "\t"
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

    def test_serialize_int(self):
        params = {
            "name": "delimiter",
                    "value": "1",
                    "type": "integer"
        }

        expected = {
            "delimiter": 1
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

    def test_serialize_int_exception(self):
        params = {
            "name": "delimiter",
                    "value": "0.1",
                    "type": "integer"
        }

        p = Parameter(**params)
        try:
            output = p.serialize()
        except ValueError as e:
            assert True

    def test_serialize_float(self):
        params = {
            "name": "delimiter",
                    "value": "0.1",
                    "type": "float"
        }

        expected = {
            "delimiter": 0.1
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

    def test_serialize_float_exception(self):
        params = {
            "name": "delimiter",
                    "value": "hello",
                    "type": "float"
        }

        p = Parameter(**params)
        try:
            output = p.serialize()
        except ValueError as e:
            assert True

    def test_serialize_bool(self):

        params = {
            "name": "delimiter",
                    "value": "true",
                    "type": "boolean"
        }

        expected = {
            "delimiter": True
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

        params = {
            "name": "delimiter",
                    "value": "True",
                    "type": "boolean"
        }

        expected = {
            "delimiter": True
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

        params = {
            "name": "delimiter",
                    "value": "false",
                    "type": "boolean"
        }

        expected = {
            "delimiter": False
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

        params = {
            "name": "delimiter",
                    "value": "False",
                    "type": "boolean"
        }

        expected = {
            "delimiter": False
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

    def test_serialize_bool_exception(self):

        params = {
            "name": "delimiter",
                    "value": "rue",
                    "type": "boolean"
        }

        p = Parameter(**params)
        try:
            output = p.serialize()
        except ValueError as e:
            assert True

    def test_dotlist_notation_to_omega_conf(self):
        """
        Tests serialized Parameter objects with dotlist notation names can be converted to OmegaConf
        which is used to construct the DataCatalog with the add_feed_dict method in tasks.py.
        """
        parameter_inputs = [{"name": "example", "value": "hello", "type": "string"},
                            {"name": "duration", "value": "0.1", "type": "float"},
                            {"name": "model_options.model_params.learning_date", "value": "2023-11-01", "type": "string"},
                            {"name": "model_options.model_params.training_date", "value": "2023-11-01", "type": "string"},
                            {"name": "model_options.model_params.data_ratio", "value": "14", "type": "float"},
                            {"name": "data_options.step_size", "value": "123123", "type": "float"},
                            ]

        parameters = [Parameter.decode(p) for p in parameter_inputs]

        serialized_parameters = {}

        for p in parameters:
            serialized_parameters.update(p.serialize())

        parameters_dotlist = [f"{key}={value}" for key, value in serialized_parameters.items()]
        conf_parameters = OmegaConf.from_dotlist(parameters_dotlist)
        kedro_parameters = {"parameters": conf_parameters}

        assert kedro_parameters == {
            'parameters': {
                'example': 'hello',
                'duration': 0.1,
                'model_options': {
                    'model_params': {
                        'learning_date': '2023-11-01',
                        'training_date': '2023-11-01',
                        'data_ratio': 14
                    }
                },
                'data_options': {
                    'step_size': 123123
                }
            }
        }

        params_dotlist = [f"params:{key}={value}" for key, value in serialized_parameters.items()]
        kedro_params = OmegaConf.from_dotlist(params_dotlist)

        assert kedro_params == {
            'params:example': 'hello',
            'params:duration': 0.1,
            'params:model_options': {
                'model_params': {
                    'learning_date': '2023-11-01',
                    'training_date': '2023-11-01',
                    'data_ratio': 14
                }
            },
            'params:data_options': {
                'step_size': 123123
            }
        }



from io import BytesIO
from uuid import uuid4

import pytest

from kedro_graphql.models import DataSet, Parameter, Pipeline, Tag
from kedro_graphql.tasks import run_pipeline

IN_DEV = True
REASON = "Argo runner in development"


@pytest.mark.skipif(IN_DEV, reason=REASON)
@pytest.fixture
def s3_client():
    from minio import Minio

    return Minio(
        "localhost:9000",
        access_key="admin",
        secret_key="password",
        secure=False
    )


@pytest.mark.skipif(IN_DEV, reason=REASON)
@pytest.fixture
def s3_object(s3_client):
    """
    https://min.io/docs/minio/linux/developers/python/API.html#put_object

    Run ```make dev-cluster``` to start minio service.
    """
    from minio.error import S3Error

    try:

        # Make 'my-bucket' bucket if not exist.
        found = s3_client.bucket_exists("my-bucket")
        if not found:
            s3_client.make_bucket("my-bucket")
        else:
            print("Bucket 'my-bucket' already exists")

        fname = "text_in_"+str(uuid4())+".txt"
        # Upload contents as object name
        # 'text_in.txt' to bucket 'my-bucket'.
        s3_client.put_object(
            "my-bucket", fname, BytesIO(b"hello"), 5,
            content_type="text"
        )
        print(
            "'"+fname+"' is successfully uploaded as "
            "object '"+fname+"' to bucket 'my-bucket'."
        )

        # https://docs.pytest.org/en/6.2.x/fixture.html#teardown-cleanup-aka-fixture-finalization
        yield "s3://my-bucket/" + fname

        # cleanup
        s3_client.remove_object("my-bucket", fname)

        print(
            "object '"+fname+"' successfully removed from "
            "bucket 'my-bucket'."
        )
    except S3Error as exc:
        print("error occurred.", exc)


@pytest.mark.skipif(IN_DEV, reason=REASON)
@pytest.fixture
def mock_pipeline_argo(mock_app, s3_object, s3_client):
    """

    """

    out_fname = "text_out_"+str(uuid4())+".txt"

    inputs = [{"name": "text_in", "type": "text.TextDataSet",
               "filepath": s3_object,
               "credentials": {"key": "admin",
                               "secret": "password",
                               "client_kwargs": {
                                   "endpoint_url": 'http://localhost:9000'
                               }}
               }]

    outputs = [{"name": "text_out", "type": "text.TextDataSet",
                "filepath": "s3://my-bucket/"+out_fname,
                "credentials": {"key": "admin",
                                "secret": "password",
                                "client_kwargs": {
                                    "endpoint_url": 'http://localhost:9000'
                                }}
                }]

    parameters = [{"name": "example", "value": "hello"}]

    tags = [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]

    p = Pipeline(
        name="example00",
        inputs=[DataSet(**i) for i in inputs],
        outputs=[DataSet(**o) for o in outputs],
        parameters=[Parameter(**p) for p in parameters],
        tags=[Tag(**p) for p in tags],
        task_name=str(run_pipeline),
    )

    serial = p.serialize()

    result = run_pipeline.apply_async(kwargs={"name": "example00",
                                              "inputs": serial["inputs"],
                                              "outputs": serial["outputs"],
                                              "parameters": serial["parameters"],
                                              "runner": "kedro_graphql.runner.argo.ArgoWorkflowsRunner"}, countdown=0.1)
    p.task_id = result.id
    p.status = result.status
    p.task_kwargs = str(
        {"name": serial["name"],
         "inputs": serial["inputs"],
         "outputs": serial["outputs"],
         "parameters": serial["parameters"]}
    )

    print(f'Starting {p.name} pipeline with task_id: ' + str(p.task_id))
    p = mock_app.backend.create(p)
    yield p
    # cleanup
    s3_client.remove_object("my-bucket", out_fname)



import logging
from importlib import import_module

logger = logging.getLogger("kedro")


def init_runner(runner=None):
    module, class_name = runner.rsplit(".", 1)

    module = import_module(module)
    runner_cls = getattr(module, class_name)
    logger.info("using runner " + str(runner_cls))
    return runner_cls



import os
import tempfile

from dotenv import dotenv_values

config = {
    "MONGO_URI": "mongodb://root:example@localhost:27017/",
    "MONGO_DB_NAME": "pipelines",
    "KEDRO_GRAPHQL_IMPORTS": "kedro_graphql.plugins.plugins,",
    "KEDRO_GRAPHQL_APP": "kedro_graphql.asgi.KedroGraphQL",
    "KEDRO_GRAPHQL_BACKEND": "kedro_graphql.backends.mongodb.MongoBackend",
    "KEDRO_GRAPHQL_BROKER": "redis://localhost",
    "KEDRO_GRAPHQL_CELERY_RESULT_BACKEND": "redis://localhost",
    "KEDRO_GRAPHQL_RUNNER": "kedro.runner.SequentialRunner",  # kedro_graphql.runner.argo.ArgoWorkflowsRunner
    "KEDRO_GRAPHQL_ENV": "local",
    "KEDRO_GRAPHQL_CONF_SOURCE": None,
    "KEDRO_GRAPHQL_DEPRECATIONS_DOCS": "https://github.com/opensean/kedro-graphql/blob/main/README.md#deprecations",
    "KEDRO_GRAPHQL_LOG_TMP_DIR": tempfile.TemporaryDirectory().name,
    "KEDRO_GRAPHQL_LOG_PATH_PREFIX": None
}

config = {
    **config,
    **dotenv_values(".env"),  # load
    **os.environ,  # override loaded values with environment variables
}



"""
This is a boilerplate pipeline 'example00'
generated using Kedro 0.18.4
"""
import time


def echo(text: str, example: str, parameters: dict) -> str:
    time.sleep(float(parameters.get("duration", 1)))
    return text
